#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.handlers.util`` -- Handler Utilities
"""

from pyramid.httpexceptions import HTTPFound

from edbob.db.perms import has_permission


class needs_perm(object):
    """
    Decorator to be used for handler methods which should restrict access based
    on the current user's permissions.
    """

    def __init__(self, permission, **kwargs):
        self.permission = permission
        self.kwargs = kwargs

    def __call__(self, fn):
        permission = self.permission
        kw = self.kwargs
        def wrapped(self):
            if not self.request.current_user:
                self.request.session['referrer'] = self.request.url_generator.current()
                self.request.session.flash("You must be logged in to do that.", 'error')
                return HTTPFound(location=self.request.route_url('login'))
            if not has_permission(self.request.current_user, permission):
                self.request.session.flash("You do not have permission to do that.", 'error')
                home = kw.get('redirect', self.request.route_url('home'))
                return HTTPFound(location=home)
            return fn(self)
        return wrapped


def needs_user(fn):
    """
    Decorator for handler methods which require simply that a user be currently
    logged in.
    """

    def wrapped(self):
        if not self.request.current_user:
            self.request.session['referrer'] = self.request.url_generator.current()
            self.request.session.flash("You must be logged in to do that.", 'error')
            return HTTPFound(location=self.request.route_url('login'))
        return fn(self)
    return wrapped



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.exceptions`` -- Exceptions
"""


class ConfigError(Exception):
    """
    Raised when configuration is missing or otherwise invalid.
    """

    def __init__(self, section, option, msg=None):
        self.section = section
        self.option = option
        self.msg = msg or "Missing or invalid config"

    def __str__(self):
        return "%s; please set '%s' in the [%s] section of your config file" % (
            self.msg, self.option, self.section)


class InitError(Exception):
    """
    Raised when initialization fails for a given module.
    """

    def __init__(self, module):
        self.module = module

    def __str__(self):
        return "Module '%s' has no init() function" % self.module.__name__


class LoadSpecError(Exception):
    """
    Raised when something obvious goes wrong with :func:`edbob.load_spec()`.
    """

    def __init__(self, spec):
        self.spec = spec

    def __str__(self):
        msg = 'Failed to load spec: %s' % self.spec
        specifics = self.specifics()
        if specifics:
            msg += " (%s)" % specifics
        return msg

    def specifics(self):
        return None


class ModuleMissingAttribute(LoadSpecError):
    """
    Raised during :func:`edbob.load_spec()` when the module imported okay but
    the attribute could not be found.
    """

    def specifics(self):
        mod, attr = self.spec.split(':')
        return "module '%s' was loaded but '%s' attribute not found" % (mod, attr)


class RecipientsNotFound(Exception):
    """
    Raised when no recipients could be found in config.
    """
    
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "No recipients configured (set 'recipients.%s' in [edbob.mail])" % self.key


class SenderNotFound(Exception):
    """
    Raised when no sender could be found in config.
    """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "No sender configured (set 'sender.%s' in [edbob.mail])" % self.key



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.filters`` -- Search Filters
"""

import re

from pyramid_simpleform import Form
from pyramid_simpleform.renderers import FormRenderer
from webhelpers.html import tags

import edbob
from edbob.util import prettify


__all__ = ['SearchFilter', 'SearchForm']


class SearchFilter(edbob.Object):

    def __init__(self, name, label=None, **kwargs):
        edbob.Object.__init__(self, **kwargs)
        self.name = name
        self.label = label or prettify(name)

    def types_select(self):
        types = [
            ('is', 'is'),
            ('nt', 'is not'),
            ('lk', 'contains'),
            ('nl', 'doesn\'t contain'),
            ]
        options = []
        filter_map = self.search.config['filter_map'][self.name]
        for value, label in types:
            if value in filter_map:
                options.append((value, label))
        return tags.select('filter_type_'+self.name,
                           self.search.config.get('filter_type_'+self.name),
                           options, class_='filter-type')

    def value_control(self):
        return tags.text(self.name, self.search.config.get(self.name))


class SearchForm(Form):

    def __init__(self, request, filters, config, *args, **kwargs):
        Form.__init__(self, request, *args, **kwargs)
        self.filters = filters
        for f in filters:
            filters[f].search = self
        self.config = config


class SearchFormRenderer(FormRenderer):

    def __init__(self, form, *args, **kwargs):
        FormRenderer.__init__(self, form, *args, **kwargs)
        self.filters = form.filters
        self.config = form.config

    def checkbox(self, name, checked=None, *args, **kwargs):
        if name.startswith('include_filter_'):
            if checked is None:
                checked = self.config[name]
            return tags.checkbox(name, checked=checked, *args, **kwargs)
        if checked is None:
            checked = False
        return FormRenderer.checkbox(self, name, checked=checked, *args, **kwargs)

    def text(self, name, *args, **kwargs):
        return tags.text(name, value=self.config.get(name), *args, **kwargs)

    def sorted_filters(self):
        return sorted(self.filters, key=lambda x: self.filters[x].label)

    def add_filter(self, visible):
        options = ['add a filter']
        for f in sorted(self.filters):
            f = self.filters[f]
            options.append((f.name, f.label))
        return self.select('add-filter', options,
                           style='display: none;' if len(visible) == len(self.filters) else None)

    def render(self, **kwargs):
        from formalchemy import config
        return config.engine('filterset', search=self, **kwargs)


def filter_exact(field):
    """
    Returns a filter map entry, with typical logic built in for "exact match"
    queries applied to ``field``.
    """
    return {
        'is':
            lambda q, v: q.filter(field == v) if v else q,
        'nt':
            lambda q, v: q.filter(field != v) if v else q,
        }


def filter_ilike(field):
    """
    Returns a filter map entry, with typical logic built in for ILIKE queries
    applied to ``field``.
    """
    return {
        'lk':
            lambda q, v: q.filter(field.ilike('%%%s%%' % v)) if v else q,
        'nl':
            lambda q, v: q.filter(~field.ilike('%%%s%%' % v)) if v else q,
        }


def filter_query(query, config, join_map={}):
    filter_map = config['filter_map']
    if config.get('search'):
        search = config['search'].config
        joins = config.setdefault('joins', [])
        include_filter = re.compile(r'^include_filter_(.*)$')
        for key in search:
            m = include_filter.match(key)
            if m and search[key]:
                field = m.group(1)
                if field in join_map and field not in joins:
                    query = join_map[field](query)
                    joins.append(field)
                value = search.get(field)
                if value:
                    f = filter_map[field][search['filter_type_'+field]]
                    query = f(query, value)
    return query


def get_filter_map(cls, exact=[], ilike=[], **kwargs):
    """
    Convenience function which returns a filter map for ``cls``.  All fields
    represented by ``names`` will be included in the map.

    Each field's entry will use the :func:`filter_ilike()` function unless the
    field's name is also found within ``exact``, in which case the
    :func:`filter_exact()` function will be used instead.
    """

    fmap = {}
    for name in exact:
        fmap[name] = filter_exact(getattr(cls, name))
    for name in ilike:
        fmap[name] = filter_ilike(getattr(cls, name))
    fmap.update(kwargs)
    return fmap


def get_search_config(name, request, filter_map, **kwargs):
    """
    Returns a dictionary of configuration options for a search form.
    """

    config = {}

    def update_config(dict_, prefix='', exclude_by_default=False):
        """
        Updates the ``config`` dictionary based on the contents of ``dict_``.
        """

        for field in filter_map:
            if prefix+'include_filter_'+field in dict_:
                include = dict_[prefix+'include_filter_'+field]
                include = bool(include) and include != '0'
                config['include_filter_'+field] = include
            elif exclude_by_default:
                config['include_filter_'+field] = False
            if prefix+'filter_type_'+field in dict_:
                config['filter_type_'+field] = dict_[prefix+'filter_type_'+field]
            if prefix+field in dict_:
                config[field] = dict_[prefix+field]

    # Update config to exclude all filters by default.
    for field in filter_map:
        config['include_filter_'+field] = False

    # Update config with defaults from ``kwargs``.
    config.update(kwargs)

    # Update config with data cached in Beaker session.
    update_config(request.session, prefix=name+'.')

    # Update config with data from GET/POST request.
    if request.params.get('filters'):
        update_config(request.params, exclude_by_default=True)

    # Cache filter data in Beaker session.
    for key in config:
        if not key.startswith('filter_factory_'):
            request.session[name+'.'+key] = config[key]

    config['request'] = request
    config['filter_map'] = filter_map
    return config


def get_search_form(config, **labels):
    filters = {}
    for field in config['filter_map']:
        factory = config.get('filter_factory_%s' % field, SearchFilter)
        filters[field] = factory(field, label=labels.get(field))
    return SearchForm(config['request'], filters, config)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.scaffolds`` -- App Scaffolding
"""

from pyramid.scaffolds import PyramidTemplate


class Template(PyramidTemplate):

    _template_dir = 'edbob'
    summary = "Pyramid edbob project"



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.grids.core`` -- Core Grid View
"""

from edbob.pyramid import grids
from edbob.pyramid.views.core import View


__all__ = ['GridView']


class GridView(View):

    route_name = None
    route_url = None
    renderer = None
    permission = None

    full = False
    checkboxes = False
    clickable = False
    deletable = False

    partial_only = False

    def update_grid_kwargs(self, kwargs):
        kwargs.setdefault('full', self.full)
        kwargs.setdefault('checkboxes', self.checkboxes)
        kwargs.setdefault('clickable', self.clickable)
        kwargs.setdefault('deletable', self.deletable)
        kwargs.setdefault('partial_only', self.partial_only)

    def make_grid(self, **kwargs):
        self.update_grid_kwargs(kwargs)
        return grids.Grid(self.request, **kwargs)

    def grid(self):
        return self.make_grid()

    def render_kwargs(self):
        return {}

    def __call__(self):
        grid = self.grid()
        kwargs = self.render_kwargs()
        return grids.util.render_grid(grid, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.users`` -- User Views
"""

from webhelpers.html import literal, tags

import formalchemy
from formalchemy.fields import SelectFieldRenderer

import edbob
from edbob.db.auth import set_user_password
from edbob.pyramid import Session
from edbob.pyramid.views import SearchableAlchemyGridView
from edbob.pyramid.views.crud import Crud


class UsersGrid(SearchableAlchemyGridView):

    mapped_class = edbob.User
    route_name = 'users'
    route_url = '/users'
    sort = 'username'

    def join_map(self):
        return {
            'person':
                lambda q: q.outerjoin(edbob.Person),
            }

    def filter_map(self):
        return self.make_filter_map(
            ilike=['username'],
            person=self.filter_ilike(edbob.Person.display_name))

    def filter_config(self):
        return self.make_filter_config(
            include_filter_username=True,
            filter_type_username='lk',
            include_filter_person=True,
            filter_type_person='lk')

    def sort_map(self):
        return self.make_sort_map(
            'username',
            person=self.sorter(edbob.Person.display_name))

    def grid(self):
        g = self.make_grid()
        g.configure(
            include=[
                g.username,
                g.person,
                ],
            readonly=True)
        return g


class _RolesFieldRenderer(SelectFieldRenderer):

    def render_readonly(self, **kwargs):
        roles = Session.query(edbob.Role)
        res = literal('<ul>')
        for uuid in self.value:
            role = roles.get(uuid)
            res += literal('<li>%s</li>' % (
                    tags.link_to(role.name,
                                 self.request.route_url('role.edit', uuid=role.uuid))))
        res += literal('</ul>')
        return res


def RolesFieldRenderer(request):
    return type('RolesFieldRenderer', (_RolesFieldRenderer,), {'request': request})


class RolesField(formalchemy.Field):

    def __init__(self, name, **kwargs):
        kwargs.setdefault('value', self.get_value)
        kwargs.setdefault('options', self.get_options())
        kwargs.setdefault('multiple', True)
        super(RolesField, self).__init__(name, **kwargs)

    def get_value(self, user):
        return [x.uuid for x in user.roles]

    def get_options(self):
        q = Session.query(edbob.Role.name, edbob.Role.uuid)
        q = q.order_by(edbob.Role.name)
        return q.all()

    def sync(self):
        if not self.is_readonly():
            user = self.model
            roles = Session.query(edbob.Role)
            data = self.renderer.deserialize()
            user.roles = [roles.get(x) for x in data]
                

class _ProtectedPersonRenderer(formalchemy.FieldRenderer):

    def render_readonly(self, **kwargs):
        res = str(self.person)
        res += tags.hidden('User--person_uuid',
                           value=self.field.parent.person_uuid.value)
        return res


def ProtectedPersonRenderer(uuid):
    person = Session.query(edbob.Person).get(uuid)
    assert person
    return type('ProtectedPersonRenderer', (_ProtectedPersonRenderer,),
                {'person': person})


class _LinkedPersonRenderer(formalchemy.FieldRenderer):

    def render_readonly(self, **kwargs):
        return tags.link_to(str(self.raw_value),
                            self.request.route_url('person.edit', uuid=self.value))


def LinkedPersonRenderer(request):
    return type('LinkedPersonRenderer', (_LinkedPersonRenderer,),
                {'request': request})


class PasswordFieldRenderer(formalchemy.PasswordFieldRenderer):

    def render(self, **kwargs):
        return tags.password(self.name, value='', maxlength=self.length, **kwargs)


def passwords_match(value, field):
    if field.parent.confirm_password.value != value:
        raise formalchemy.ValidationError("Passwords do not match")
    return value


class PasswordField(formalchemy.Field):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('value', lambda x: x.password)
        kwargs.setdefault('renderer', PasswordFieldRenderer)
        kwargs.setdefault('validate', passwords_match)
        super(PasswordField, self).__init__(*args, **kwargs)

    def sync(self):
        if not self.is_readonly():
            password = self.renderer.deserialize()
            if password:
                set_user_password(self.model, password)


class UserCrud(Crud):

    mapped_class = edbob.User
    home_route = 'users.list'

    def fieldset(self, user):
        fs = self.make_fieldset(user)

        fs.append(PasswordField('password'))
        fs.append(formalchemy.Field('confirm_password',
                                    renderer=PasswordFieldRenderer))
        fs.append(RolesField('roles',
                             renderer=RolesFieldRenderer(self.request)))

        fs.configure(
            include=[
                fs.username,
                fs.person,
                fs.password.label("Set Password"),
                fs.confirm_password,
                fs.roles,
                ])

        # if fs.edit and user.person:
        if isinstance(user, edbob.User) and user.person:
            fs.person.set(readonly=True,
                          renderer=LinkedPersonRenderer(self.request))

        return fs

    def validation_failed(self, fs):
        if not fs.edit and fs.person_uuid.value:
            fs.person.set(readonly=True,
                          renderer=ProtectedPersonRenderer(fs.person_uuid.value))


def includeme(config):
    UsersGrid.add_route(config)
    UserCrud.add_routes(config)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.model`` -- Core Schema Definition
"""

from sqlalchemy import Column, String, Text

import edbob
from edbob.db import Base


__all__ = ['ActiveExtension', 'Setting']


def uuid_column(*args):
    """
    Convenience function which returns a ``uuid`` column for use as a table's
    primary key.
    """

    return Column(String(32), primary_key=True, default=edbob.get_uuid, *args)


class ActiveExtension(Base):
    """
    Represents an extension which has been activated within a database.
    """

    __tablename__ = 'active_extensions'

    name = Column(String(50), primary_key=True)

    def __repr__(self):
        return "<ActiveExtension: %s>" % self.name

    def __str__(self):
        return str(self.name or '')


class Setting(Base):
    """
    Represents a setting stored within the database.
    """

    __tablename__ = 'settings'

    name = Column(String(255), primary_key=True)
    value = Column(Text)

    def __repr__(self):
        return "<Setting: %s>" % self.name



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.auth`` -- Authentication & Authorization
"""

import bcrypt

from sqlalchemy.orm import object_session

import edbob
from edbob.db import needs_session


class BcryptAuthenticator(edbob.Object):
    """
    Authentication with py-bcrypt (Blowfish).
    """

    def populate_user(self, user, password):
        user.salt = bcrypt.gensalt()
        user.password = bcrypt.hashpw(password, user.salt)

    def authenticate_user(self, user, password):
        return bcrypt.hashpw(password, user.salt) == user.password


@needs_session
def authenticate_user(session, username, password):
    """
    Attempts to authenticate with ``username`` and ``password``.  If successful,
    returns the :class:`edbob.User` instance; otherwise returns ``None``.
    """

    q = session.query(edbob.User)
    q = q.filter(edbob.User.username == username)
    user = q.first()
    if user:
        auth = BcryptAuthenticator()
        if auth.authenticate_user(user, password):
            return user


def administrator_role(session):
    """
    Returns the "Administrator" :class:`edbob.Role` instance, attached to the
    given ``session``.
    """

    uuid = 'd937fa8a965611dfa0dd001143047286'
    admin = session.query(edbob.Role).get(uuid)
    if admin:
        return admin
    admin = edbob.Role(uuid=uuid, name='Administrator')
    session.add(admin)
    return admin


def guest_role(session):
    """
    Returns the "Guest" :class:`edbob.Role` instance, attached to the given
    ``session``.
    """

    uuid = 'f8a27c98965a11dfaff7001143047286'
    guest = session.query(edbob.Role).get(uuid)
    if guest:
        return guest
    guest = edbob.Role(uuid=uuid, name='Guest')
    session.add(guest)
    return guest


def grant_permission(role, permission, session=None):
    """
    Grants ``permission`` to ``role``.
    """

    if not session:
        session = object_session(role)
        assert session
    if permission not in role.permissions:
        role.permissions.append(permission)


def has_permission(obj, perm, session=None):
    """
    Checks the given ``obj`` (which may be either a :class:`edbob.User`` or
    :class:`edbob.Role` instance), and returns a boolean indicating whether or
    not the object is allowed the given permission.  ``perm`` should be a
    fully-qualified permission name, e.g. ``'users.create'``.
    """

    if isinstance(obj, edbob.User):
        roles = list(obj.roles)
    elif isinstance(obj, edbob.Role):
        roles = [obj]
    elif obj is None:
        roles = []
    else:
        raise TypeError("You must pass either a User or Role for 'obj'; got: %s" % repr(obj))
    if not session:
        session = object_session(obj)
        assert session
    admin = administrator_role(session)
    roles.append(guest_role(session))
    for role in roles:
        if role is admin:
            return True
        for permission in role.permissions:
            if permission == perm:
                return True
    return False


def init_database(engine, session):
    """
    Initialize the auth system within an ``edbob`` database.

    Currently this only creates an :class:`edbob.User` instance with username
    ``'admin'`` (and password the same), and assigns the user to the built-in
    administrative role (see :func:`administrator_role()`).
    """

    admin = edbob.User(username='admin')
    set_user_password(admin, 'admin')
    admin.roles.append(administrator_role(session))
    session.add(admin)
    session.flush()
    print "Created 'admin' user with password 'admin'"


def set_user_password(user, password):
    """
    Sets the password for the given :class:`edbob.User` instance.
    """

    auth = BcryptAuthenticator()
    auth.populate_user(user, password)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.util`` -- Utilities
"""

from pkg_resources import iter_entry_points

import edbob


def entry_point_map(key):
    """
    Convenience function to retrieve a dictionary of entry points, keyed by
    name.

    ``key`` must be the "section name" for the entry points you're after, e.g.
    ``'edbob.commands'``.
    """

    epmap = {}
    for ep in iter_entry_points(key):
        epmap[ep.name] = ep.load()
    return epmap


def prettify(text):
    """
    Returns a "prettified" version of ``text``, which is more or less assumed
    to be a Pythonic representation of an (singular or plural) entity name.  It
    splits the text into capitalized words, e.g. "purchase_orders" becomes
    "Purchase Orders".

    .. note::
       No attempt is made to handle pluralization; the spelling of ``text`` is
       always preserved.
    """

    words = text.replace('_', ' ').split()
    return ' '.join([x.capitalize() for x in words])


class requires_impl(edbob.Object):
    """
    Decorator for properties or methods defined on parent classes only for
    documentation's sake, but which in fact rely on the derived class entirely
    for implementation.

    This merely adds a helpful message to the ``NotImplementedError`` exception
    which will be raised.
    """

    is_property = False

    def __call__(self, func):
        if self.is_property:
            message = "Please define the %s.%s attribute"
        else:
            message = "Please implement the %s.%s() method"

        def wrapped(self, *args, **kwargs):
            msg = message % (self.__class__.__name__, func.__name__)
            msg += " (within the %s module)" % self.__class__.__module__
            raise NotImplementedError(msg)

        # wrapped.__doc__ = func.__doc__ + "\n\n    This must be implemented by the derived class."
        wrapped.__doc__ = """%s

        This must be implemented by the derived class.""" % func.__doc__
        return wrapped



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.util`` -- Database Utilities
"""

import sqlalchemy.exc

import edbob.db
from edbob.db.model import ActiveExtension


def core_schema_installed(engine=None):
    """
    Returns boolean indicating whether the core schema has been installed to
    the database represented by ``engine``.

    If no engine is provided, :attr:`edbob.db.engine` is assumed.
    """
    
    if engine is None:
        engine = edbob.db.engine

    # Check database existence and/or connectivity.
    try:
        conn = engine.connect()
    except sqlalchemy.exc.OperationalError:
        return False
    else:
        conn.close()

    # Issue "bogus" query to verify core table existence.
    session = edbob.db.Session(bind=engine)
    try:
        session.query(ActiveExtension).count()
    except sqlalchemy.exc.ProgrammingError:
        return False
    finally:
        session.close()

    return True


def install_core_schema(engine=None):
    """
    Installs the core schema to the database represented by ``engine``.

    If no engine is provided, :attr:`edbob.db.engine` is assumed.
    """

    if not engine:
        engine = edbob.db.engine

    # Attempt connection in order to force an error, if applicable.
    conn = engine.connect()
    conn.close()

    # Create tables for core schema.
    meta = edbob.db.get_core_metadata()
    meta.create_all(engine)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.grids.alchemy`` -- FormAlchemy Grid Classes
"""

from webhelpers.html import tags
from webhelpers.html import HTML

import formalchemy

import edbob
from edbob.pyramid import Session
from edbob.pyramid.grids.core import Grid
from edbob.util import prettify


__all__ = ['AlchemyGrid']


class AlchemyGrid(Grid):

    sort_map = {}

    pager = None
    pager_format = '$link_first $link_previous ~1~ $link_next $link_last'

    def __init__(self, request, cls, instances, **kwargs):
        super(AlchemyGrid, self).__init__(request, **kwargs)
        self._formalchemy_grid = formalchemy.Grid(
            cls, instances, session=Session(), request=request)
        self._formalchemy_grid.prettify = prettify
        self.noclick_fields = []

    def __getattr__(self, attr):
        return getattr(self._formalchemy_grid, attr)

    def cell_class(self, field):
        classes = [field.name]
        if field.name in self.noclick_fields:
            classes.append('noclick')
        return ' '.join(classes)

    def checkbox(self, row):
        return tags.checkbox('check-'+row.uuid)

    def click_route_kwargs(self, row):
        return {'uuid': row.uuid}

    def column_header(self, field):
        class_ = None
        label = field.label()
        if field.key in self.sort_map:
            class_ = 'sortable'
            if field.key == self.config['sort']:
                class_ += ' sorted ' + self.config['dir']
            label = tags.link_to(label, '#')
        return HTML.tag('th', class_=class_, field=field.key,
                        title=self.column_titles.get(field.key), c=label)

    def edit_route_kwargs(self, row):
        return {'uuid': row.uuid}

    def delete_route_kwargs(self, row):
        return {'uuid': row.uuid}

    def iter_fields(self):
        return self._formalchemy_grid.render_fields.itervalues()

    def iter_rows(self):
        for row in self._formalchemy_grid.rows:
            self._formalchemy_grid._set_active(row)
            yield row

    def page_count_options(self):
        options = edbob.config.get('edbob.pyramid', 'grid.page_count_options')
        if options:
            options = options.split(',')
            options = [int(x.strip()) for x in options]
        else:
            options = [5, 10, 20, 50, 100]
        return options

    def page_links(self):
        return self.pager.pager(self.pager_format,
                                symbol_next='next',
                                symbol_previous='prev',
                                onclick="grid_navigate_page(this, '$partial_url'); return false;")

    def render_field(self, field):
        if self._formalchemy_grid.readonly:
            return field.render_readonly()
        return field.render()

    def row_attrs(self, row, i):
        attrs = super(AlchemyGrid, self).row_attrs(row, i)
        if hasattr(row, 'uuid'):
            attrs['uuid'] = row.uuid
        return attrs



#!/usr/bin/env python
# -*- coding: utf-8  -*-

"""
pretty

Formats dates, numbers, etc. in a pretty, human readable format.
"""
__author__ = "S Anand (sanand@s-anand.net)"
__copyright__ = "Copyright 2010, S Anand"
__license__ = "WTFPL"


# Note that some modifications exist between this file and the original as
# downloaded from http://pypi.python.org/pypi/py-pretty


from datetime import datetime

import edbob


def _df(seconds, denominator=1, text='', past=True):
    if past:   return         str((seconds + denominator/2)/ denominator) + text + ' ago'
    else:      return 'in ' + str((seconds + denominator/2)/ denominator) + text


def date(time=False, asdays=False, short=False):
    '''Returns a pretty formatted date.
    Inputs:
        time is a datetime object or an int timestamp
        asdays is True if you only want to measure days, not seconds
        short is True if you want "1d ago", "2d ago", etc. False if you want
    '''

    now = edbob.utc_time()
    if type(time) is int:   time = datetime.fromtimestamp(time)
    elif not time:          time = now

    if time > now:  past, diff = False, time - now
    else:           past, diff = True,  now - time
    seconds = diff.seconds
    days    = diff.days

    if short:
        if days == 0 and not asdays:
            if   seconds < 10:          return 'now'
            elif seconds < 60:          return _df(seconds, 1, 's', past)
            elif seconds < 3600:        return _df(seconds, 60, 'm', past)
            else:                       return _df(seconds, 3600, 'h', past)
        else:
            if   days   == 0:           return 'today'
            elif days   == 1:           return past and 'yest' or 'tom'
            elif days    < 7:           return _df(days, 1, 'd', past)
            elif days    < 31:          return _df(days, 7, 'w', past)
            elif days    < 365:         return _df(days, 30, 'mo', past)
            else:                       return _df(days, 365, 'y', past)
    else:
        if days == 0 and not asdays:
            if   seconds < 10:          return 'now'
            elif seconds < 60:          return _df(seconds, 1, ' seconds', past)
            elif seconds < 120:         return past and 'a minute ago' or 'in a minute'
            elif seconds < 3600:        return _df(seconds, 60, ' minutes', past)
            elif seconds < 7200:        return past and 'an hour ago' or'in an hour'
            else:                       return _df(seconds, 3600, ' hours', past)
        else:
            if   days   == 0:           return 'today'
            elif days   == 1:           return past and 'yesterday' or'tomorrow'
            elif days   == 2:           return past and 'day before' or 'day after'
            elif days    < 7:           return _df(days, 1, ' days', past)
            elif days    < 14:          return past and 'last week' or 'next week'
            elif days    < 31:          return _df(days, 7, ' weeks', past)
            elif days    < 61:          return past and 'last month' or 'next month'
            elif days    < 365:         return _df(days, 30, ' months', past)
            elif days    < 730:         return past and 'last year' or 'next year'
            else:                       return _df(days, 365, ' years', past)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.extensions.contact.model`` -- Schema Definition
"""

from sqlalchemy import Column, String, Integer
from sqlalchemy import and_
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

from edbob.db.model import Base, uuid_column


__all__ = ['PhoneNumber', 'Person', 'PersonPhoneNumber', 'EmailAddress',
           'PersonEmailAddress']


def get_person_display_name(context):
    """
    Provides a default value for :attr:`Person.display_name`, constructed from
    :attr:`Person.first_name` and :attr:`Person.last_name`.
    """

    first_name = context.current_parameters['first_name']
    last_name = context.current_parameters['last_name']
    if first_name and last_name:
        return first_name + ' ' + last_name
    if first_name:
        return first_name
    if last_name:
        return last_name
    return None


class PhoneNumber(Base):
    """
    Represents a phone (or fax) number associated with a contactable entity.
    """

    __tablename__ = 'phone_numbers'

    uuid = uuid_column()
    parent_type = Column(String(20), nullable=False)
    parent_uuid = Column(String(32), nullable=False)
    preference = Column(Integer, nullable=False)
    type = Column(String(15))
    number = Column(String(20), nullable=False)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.number)

    def __unicode__(self):
        return unicode(self.number)


class PersonPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a :class:`Person`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Person'}


class EmailAddress(Base):
    """
    Represents an email address associated with a contactable entity.
    """

    __tablename__ = 'email_addresses'

    uuid = uuid_column()
    parent_type = Column(String(20), nullable=False)
    parent_uuid = Column(String(32), nullable=False)
    preference = Column(Integer, nullable=False)
    type = Column(String(15))
    address = Column(String(255), nullable=False)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.address)

    def __unicode__(self):
        return unicode(self.address)


class PersonEmailAddress(EmailAddress):
    """
    Represents an email address associated with a :class:`Person`.
    """

    __mapper_args__ = {'polymorphic_identity': 'Person'}


class Person(Base):
    """
    Represents a real, living and breathing person.  (Or, at least was
    previously living and breathing, in the case of the deceased.)
    """

    __tablename__ = 'people'

    uuid = uuid_column()
    first_name = Column(String(50))
    last_name = Column(String(50))
    display_name = Column(String(100), default=get_person_display_name)

    def __repr__(self):
        return "<Person: %s>" % self.display_name

    def __unicode__(self):
        return unicode(self.display_name or '')

    def add_email_address(self, address, type='Home'):
        email = PersonEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Home'):
        phone = PersonPhoneNumber(number=number, type=type)
        self.phones.append(phone)

Person.emails = relationship(
    PersonEmailAddress,
    backref='person',
    primaryjoin=PersonEmailAddress.parent_uuid == Person.uuid,
    foreign_keys=[PersonEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=PersonEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Person.email = relationship(
    PersonEmailAddress,
    primaryjoin=and_(
        PersonEmailAddress.parent_uuid == Person.uuid,
        PersonEmailAddress.preference == 1,
        ),
    foreign_keys=[PersonEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)

Person.phones = relationship(
    PersonPhoneNumber,
    backref='person',
    primaryjoin=PersonPhoneNumber.parent_uuid == Person.uuid,
    foreign_keys=[PersonPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=PersonPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Person.phone = relationship(
    PersonPhoneNumber,
    primaryjoin=and_(
        PersonPhoneNumber.parent_uuid == Person.uuid,
        PersonPhoneNumber.preference == 1,
        ),
    foreign_keys=[PersonPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.grids`` -- Grid Views
"""

from webhelpers import paginate

from edbob.pyramid import grids
from edbob.pyramid import Session
from edbob.pyramid.views.grids.core import GridView
from edbob.util import requires_impl


__all__ = ['AlchemyGridView', 'SortableAlchemyGridView',
           'PagedAlchemyGridView', 'SearchableAlchemyGridView']


class AlchemyGridView(GridView):

    def make_query(self):
        q = Session.query(self.mapped_class)
        return q

    def query(self):
        return self.make_query()

    def make_grid(self, **kwargs):
        self.update_grid_kwargs(kwargs)
        return grids.AlchemyGrid(
            self.request, self.mapped_class, self._data, **kwargs)

    def grid(self):
        return self.make_grid()

    def __call__(self):
        self._data = self.query()
        grid = self.grid()
        return grids.util.render_grid(grid)


class SortableAlchemyGridView(AlchemyGridView):

    sort = None

    @property
    @requires_impl(is_property=True)
    def config_prefix(self):
        pass

    def join_map(self):
        return {}

    def make_sort_map(self, *args, **kwargs):
        return grids.util.get_sort_map(
            self.mapped_class, names=args or None, **kwargs)

    def sorter(self, field):
        return grids.util.sorter(field)

    def sort_map(self):
        return self.make_sort_map()

    def make_sort_config(self, **kwargs):
        return grids.util.get_sort_config(
            self.config_prefix, self.request, **kwargs)

    def sort_config(self):
        return self.make_sort_config(sort=self.sort)

    def make_query(self):
        query = Session.query(self.mapped_class)
        query = grids.util.sort_query(
            query, self._sort_config, self.sort_map(), self.join_map())
        return query

    def query(self):
        return self.make_query()

    def make_grid(self, **kwargs):
        self.update_grid_kwargs(kwargs)
        return grids.AlchemyGrid(
            self.request, self.mapped_class, self._data,
            sort_map=self.sort_map(), config=self._sort_config, **kwargs)

    def grid(self):
        return self.make_grid()

    def __call__(self):
        self._sort_config = self.sort_config()
        self._data = self.query()
        grid = self.grid()
        return grids.util.render_grid(grid)


class PagedAlchemyGridView(SortableAlchemyGridView):

    full = True

    def make_pager(self):
        config = self._sort_config
        query = self.query()
        return paginate.Page(
            query, item_count=query.count(),
            items_per_page=int(config['per_page']),
            page=int(config['page']),
            url=paginate.PageURL_WebOb(self.request))

    def __call__(self):
        self._sort_config = self.sort_config()
        self._data = self.make_pager()
        grid = self.grid()
        grid.pager = self._data
        return grids.util.render_grid(grid)


class SearchableAlchemyGridView(PagedAlchemyGridView):

    def filter_exact(self, field):
        return grids.search.filter_exact(field)

    def filter_ilike(self, field):
        return grids.search.filter_ilike(field)

    def make_filter_map(self, **kwargs):
        return grids.search.get_filter_map(self.mapped_class, **kwargs)

    def filter_map(self):
        return self.make_filter_map()

    def make_filter_config(self, **kwargs):
        return grids.search.get_filter_config(
            self.config_prefix, self.request, self.filter_map(), **kwargs)

    def filter_config(self):
        return self.make_filter_config()

    def make_search_form(self):
        return grids.search.get_search_form(
            self.request, self.filter_map(), self._filter_config)

    def search_form(self):
        return self.make_search_form()

    def make_query(self, session=Session):
        join_map = self.join_map()
        query = session.query(self.mapped_class)
        query = grids.search.filter_query(
            query, self._filter_config, self.filter_map(), join_map)
        if hasattr(self, '_sort_config'):
            self._sort_config['joins'] = self._filter_config['joins']
            query = grids.util.sort_query(
                query, self._sort_config, self.sort_map(), join_map)
        return query

    def __call__(self):
        self._filter_config = self.filter_config()
        search = self.search_form()
        self._sort_config = self.sort_config()
        self._data = self.make_pager()
        grid = self.grid()
        grid.pager = self._data
        kwargs = self.render_kwargs()
        return grids.util.render_grid(grid, search, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.modules`` -- Module Tools
"""

import sys

from edbob import exceptions


__all__ = ['load_spec']


def import_module_path(module_path):
    """
    Imports and returns an arbitrary Python module, given its "dotted" path
    (i.e. not its file path).
    """

    if module_path in sys.modules:
        return sys.modules[module_path]
    module = __import__(module_path)
    return last_module(module, module_path)


def last_module(module, module_path):
    """
    Returns the "last" module represented by ``module_path``, by walking
    ``module`` until the desired module is found.

    For example, passing a reference to the ``rattail`` module and a module
    path of ``"rattail.sw.ishida.slpv"``, this function will ultimately return
    a reference to the actual ``rattail.sw.ishida.slpv`` module.

    This function is primarily used by :func:`import_module_path()`, since
    Python's ``__import__()`` function will typically return the top-most
    ("first") module in the dotted path.
    """

    parts = module_path.split('.')
    parts.pop(0)
    child = getattr(module, parts[0])
    if len(parts) == 1:
        return child
    return last_module(child, '.'.join(parts))


def load_spec(spec):
    """
    .. highlight:: none

    Returns an object as found in a module namespace.  ``spec`` should be of
    the same form which setuptools uses for its entry points, e.g.::

       rattail.ce:collect_batch

    The above would return a reference to the ``collect_batch`` function found
    in the ``rattail.ce`` module namespace.  The module is loaded (imported) if
    necessary.
    """

    module_path, obj = spec.split(':')
    module = import_module_path(module_path)
    try:
        obj = getattr(module, obj)
    except AttributeError:
        raise exceptions.ModuleMissingAttribute(spec)
    return obj



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.mappers`` -- Object Relational Mappings
"""

from sqlalchemy.orm import mapper, relationship

from edbob.db import classes as c


def make_mappers(metadata):
    """
    This function glues together the schema definition found in the ``models``
    module with the data class definitions found in the ``classes`` module.
    ``metadata`` should be a ``sqlalchemy.MetaData`` instance.

    It is meant to be called only once, by :func:`edbob.init()`.
    """

    t = metadata.tables


    # ActiveExtension
    mapper(
        c.ActiveExtension, t['active_extensions'],
        )


    # Permission
    mapper(
        c.Permission, t['permissions'],
        )


    # Person
    mapper(
        c.Person, t['people'],
        properties=dict(

            user=relationship(
                c.User,
                back_populates='person',
                uselist=False,
                ),
            ),
        )


    # Role
    mapper(
        c.Role, t['roles'],
        properties=dict(

            _permissions=relationship(
                c.Permission,
                backref='role',
                ),

            _users=relationship(
                c.UserRole,
                backref='role',
                ),
            ),
        )


    # Setting
    mapper(
        c.Setting, t['settings'],
        )


    # User
    mapper(
        c.User, t['users'],
        properties=dict(

            person=relationship(
                c.Person,
                back_populates='user',
                ),

            _roles=relationship(
                c.UserRole,
                backref='user',
                cascade='save-update,merge,delete',
                ),
            ),
        )


    # UserRole
    mapper(
        c.UserRole, t['users_roles'],
        )



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.mail`` -- Email Framework
"""

import logging
import smtplib
from email.message import Message

import edbob
from edbob import exceptions


log = logging.getLogger(__name__)


def sendmail(sender, recipients, subject, body, content_type='text/plain'):
    """
    Sends an email message from ``sender`` to each address in ``recipients``
    (which should be a sequence), with subject ``subject`` and body ``body``.

    If sending an HTML message instead of plain text, be sure to set
    ``content_type`` to ``'text/html'``.
    """
    message = Message()
    message.set_type(content_type)
    message['From'] = sender
    for recipient in recipients:
        message['To'] = recipient
    message['Subject'] = subject
    message.set_payload(body)

    server = edbob.config.get('edbob.mail', 'smtp.server',
                              default='localhost')
    username = edbob.config.get('edbob.mail', 'smtp.username')
    password = edbob.config.get('edbob.mail', 'smtp.password')

    log.debug("sendmail: connecting to server: %s" % server)
    session = smtplib.SMTP(server)
    if username and password:
        res = session.login(username, password)
        log.debug("sendmail: login result is: %s" % str(res))
    res = session.sendmail(message['From'], message.get_all('To'), message.as_string())
    log.debug("sendmail: sendmail result is: %s" % res)
    session.quit()


def get_sender(key):
    sender = edbob.config.get('edbob.mail', 'sender.'+key)
    if sender:
        return sender
    sender = edbob.config.get('edbob.mail', 'sender.default')
    if sender:
        return sender
    raise exceptions.SenderNotFound(key)


def get_recipients(key):
    recips = edbob.config.get('edbob.mail', 'recipients.'+key)
    if recips:
        return eval(recips)
    recips = edbob.config.get('edbob.mail', 'recipients.default')
    if recips:
        return eval(recips)
    raise exceptions.RecipientsNotFound(key)


def get_subject(key):
    subject = edbob.config.get('edbob.mail', 'subject.'+key)
    if subject:
        return subject
    subject = edbob.config.get('edbob.mail', 'subject.default')
    if subject:
        return subject
    return "[edbob]"


def sendmail_with_config(key, body, subject=None, **kwargs):
    """
    .. highlight:: ini

    Sends mail using sender/recipient/subject values found in config, according
    to ``key``.  Probably the easiest way to explain would be to show an example::
    
       [edbob.mail]
       smtp.server = localhost
       sender.default = Lance Edgar <lance@edbob.org>
       subject.default = A Nice Shrubbery, Not Too Expensive

       recipients.nightly_reports = ['Lance Edgar <lance@edbob.org>']

       subject.tragic_errors = The World Is Nearing The End!!
       recipients.tragic_errors = ['Lance Edgar <lance@edbob.org>']

    Anything not configured explicitly will fall back to defaults where
    possible.  Note however that a sender and recipients (default or otherwise)
    *must* be found or else an exception will be raised.

    The above does not include a default recipient list, but it would work the same
    as the subject and sender as far as the key goes.  To send these mails then::

       from edbob.mail import sendmail_with_config

       # just a report
       sendmail_with_config('nightly_reports', open('report.txt').read())

       # you might want to sit down for this one...
       sendmail_with_config('tragic_errors', open('OMGWTFBBQ.txt').read())

    If you do not provide ``subject`` to this function, and there is no
    ``subject.default`` setting found in config, a default of "[edbob]" will
    be used.
    """
    if subject is None:
        subject = get_subject(key)
    return sendmail(get_sender(key), get_recipients(key), subject, body, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.handlers.base`` -- Base Handlers
"""

from pyramid.renderers import render_to_response
from pyramid.httpexceptions import HTTPException, HTTPFound, HTTPOk, HTTPUnauthorized

# import sqlahelper

# # import rattail.pyramid.forms.util as util
# from rattail.db.perms import has_permission
# from rattail.pyramid.forms.formalchemy import Grid


class needs_perm(object):
    """
    Decorator to be used for handler methods which should restrict access based
    on the current user's permissions.
    """

    def __init__(self, permission, **kwargs):
        self.permission = permission
        self.kwargs = kwargs

    def __call__(self, fn):
        permission = self.permission
        kw = self.kwargs
        def wrapped(self):
            if not self.request.current_user:
                self.request.session['referrer'] = self.request.url_generator.current()
                self.request.session.flash("You must be logged in to do that.", 'error')
                return HTTPFound(location=self.request.route_url('login'))
            if not has_permission(self.request.current_user, permission):
                self.request.session.flash("You do not have permission to do that.", 'error')
                home = kw.get('redirect', self.request.route_url('home'))
                return HTTPFound(location=home)
            return fn(self)
        return wrapped


def needs_user(fn):
    """
    Decorator for handler methods which require simply that a user be currently
    logged in.
    """

    def wrapped(self):
        if not self.request.current_user:
            self.request.session['referrer'] = self.request.url_generator.current()
            self.request.session.flash("You must be logged in to do that.", 'error')
            return HTTPFound(location=self.request.route_url('login'))
        return fn(self)
    return wrapped


class Handler(object):

    def __init__(self, request):
        self.request = request
        self.Session = sqlahelper.get_session()

    # def json_response(self, data={}):
    #     response = render_to_response('json', data, request=self.request)
    #     response.headers['Content-Type'] = 'application/json'
    #     return response


class CrudHandler(Handler):
    # """
    # This handler provides all the goodies typically associated with general
    # CRUD functionality, e.g. search filters and grids.
    # """

    def crud(self, cls, fieldset_factory, home=None, delete=None, post_sync=None, pre_render=None):
        """
        Adds a common CRUD mechanism for objects.

        ``cls`` should be a SQLAlchemy-mapped class, presumably deriving from
        :class:`rattail.Object`.

        ``fieldset_factory`` must be a callable which accepts the fieldset's
        "model" as its only positional argument.

        ``home`` will be used as the redirect location once a form is fully
        validated and data saved.  If you do not speficy this parameter, the
        user will be redirected to be the CRUD page for the new object (e.g. so
        an object may be created before certain properties may be edited).

        ``delete`` may either be a string containing a URL to which the user
        should be redirected after the object has been deleted, or else a
        callback which will be executed *instead of* the normal algorithm
        (which is merely to delete the object via the Session).

        ``post_sync`` may be a callback which will be executed immediately
        after ``FieldSet.sync()`` is called, i.e. after validation as well.

        ``pre_render`` may be a callback which will be executed after any POST
        processing has occured, but just before rendering.
        """

        uuid = self.request.params.get('uuid')
        obj = self.Session.query(cls).get(uuid) if uuid else cls
        assert obj

        if self.request.params.get('delete'):
            if delete:
                if isinstance(delete, basestring):
                    self.Session.delete(obj)
                    return HTTPFound(location=delete)
                res = delete(obj)
                if res:
                    return res
            else:
                self.Session.delete(obj)
                if not home:
                    raise ValueError("Must specify 'home' or 'delete' url "
                                     "in call to CrudHandler.crud()")
                return HTTPFound(location=home)

        fs = fieldset_factory(obj)

        # if not fs.readonly and self.request.params.get('fieldset'):
        #     fs.rebind(data=self.request.params)
        #     if fs.validate():
        #         fs.sync()
        #         if post_sync:
        #             res = post_sync(fs)
        #             if isinstance(res, HTTPFound):
        #                 return res
        #         if self.request.params.get('partial'):
        #             self.Session.flush()
        #             return self.json_success(uuid=fs.model.uuid)
        #         return HTTPFound(location=self.request.route_url(objects, action='index'))

        if not fs.readonly and self.request.POST:
            # print self.request.POST
            fs.rebind(data=self.request.params)
            if fs.validate():
                fs.sync()
                if post_sync:
                    res = post_sync(fs)
                    if res:
                        return res
                if self.request.params.get('partial'):
                    self.Session.flush()
                    return self.json_success(uuid=fs.model.uuid)

                if not home:
                    self.Session.flush()
                    home = self.request.url_generator.current() + '?uuid=' + fs.model.uuid
                    self.request.session.flash("%s \"%s\" has been %s." % (
                            fs.crud_title, fs.get_display_text(),
                            'updated' if fs.edit else 'created'))
                return HTTPFound(location=home)

        data = {'fieldset': fs, 'crud': True}

        if pre_render:
            res = pre_render(fs)
            if res:
                if isinstance(res, HTTPException):
                    return res
                data.update(res)

        # data = {'fieldset':fs}
        # if self.request.params.get('partial'):
        #     return render_to_response('/%s/crud_partial.mako' % objects,
        #                               data, request=self.request)
        # return data

        return data

    def grid(self, *args, **kwargs):
        """
        Convenience function which returns a grid.  The only functionality this
        method adds is the ``session`` parameter.
        """

        return Grid(session=self.Session(), *args, **kwargs)

    # def get_grid(self, name, grid, query, search=None, url=None, **defaults):
    #     """
    #     Convenience function for obtaining the configuration for a grid,
    #     and then obtaining the grid itself.

    #     ``name`` is essentially the config key, e.g. ``'products.lookup'``, and
    #     in fact is expected to take that precise form (where the first part is
    #     considered the handler name and the second part the action name).

    #     ``grid`` must be a callable with a signature of ``grid(query,
    #     config)``, and ``query`` will be passed directly to the ``grid``
    #     callable.  ``search`` will be used to inform the grid of the search in
    #     effect, if any. ``defaults`` will be used to customize the grid config.
    #     """

    #     if not url:
    #         handler, action = name.split('.')
    #         url = self.request.route_url(handler, action=action)
    #     config = util.get_grid_config(name, self.request, search,
    #                                   url=url, **defaults)
    #     return grid(query, config)

    # def get_search_form(self, name, labels={}, **defaults):
    #     """
    #     Convenience function for obtaining the configuration for a search form,
    #     and then obtaining the form itself.

    #     ``name`` is essentially the config key, e.g. ``'products.lookup'``.
    #     The ``labels`` dictionary can be used to override the default labels
    #     displayed for the various search fields.  The ``defaults`` dictionary
    #     is used to customize the search config.
    #     """

    #     config = util.get_search_config(name, self.request,
    #                                     self.filter_map(), **defaults)
    #     form = util.get_search_form(config, **labels)
    #     return form

    # def object_crud(self, cls, objects=None, post_sync=None):
    #     """
    #     This method is a desperate attempt to encapsulate shared CRUD logic
    #     which is useful across all editable data objects.

    #     ``objects``, if provided, should be the plural name for the class as
    #     used in internal naming, e.g. ``'products'``.  A default will be used
    #     if you do not provide this value.

    #     ``post_sync``, if provided, should be a callable which accepts a
    #     ``formalchemy.Fieldset`` instance as its only argument.  It will be
    #     called immediately after the fieldset is synced.
    #     """

    #     if not objects:
    #         objects = cls.__name__.lower() + 's'

    #     uuid = self.request.params.get('uuid')
    #     obj = self.Session.query(cls).get(uuid) if uuid else cls
    #     assert obj

    #     fs = self.fieldset(obj)

    #     if not fs.readonly and self.request.params.get('fieldset'):
    #         fs.rebind(data=self.request.params)
    #         if fs.validate():
    #             fs.sync()
    #             if post_sync:
    #                 res = post_sync(fs)
    #                 if isinstance(res, HTTPFound):
    #                     return res
    #             if self.request.params.get('partial'):
    #                 self.Session.flush()
    #                 return self.json_success(uuid=fs.model.uuid)
    #             return HTTPFound(location=self.request.route_url(objects, action='index'))

    #     data = {'fieldset':fs}
    #     if self.request.params.get('partial'):
    #         return render_to_response('/%s/crud_partial.mako' % objects,
    #                                   data, request=self.request)
    #     return data

    # def render_grid(self, grid, search=None, **kwargs):
    #     """
    #     Convenience function to render a standard grid.  Really just calls
    #     :func:`dtail.forms.util.render_grid()`.
    #     """

    #     return util.render_grid(self.request, grid, search, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.initialization`` -- Initialization Framework
"""

import os
import logging

import edbob
from edbob.configuration import (
    AppConfigParser,
    default_system_paths,
    default_user_paths,
    )
from edbob.exceptions import InitError


__all__ = ['init', 'init_modules', 'inited']

inited = []

log = logging.getLogger(__name__)


def init(appname='edbob', *args, **kwargs):
    """
    Initializes the edbob framework, typically by first reading some config
    file(s) to determine which interfaces to engage.  This must normally be
    called prior to doing anything really useful, as it is responsible for
    extending the live API in-place.

    The meaning of ``args`` is as follows:

    If ``args`` is empty, the ``EDBOB_CONFIG`` environment variable is first
    consulted.  If it is nonempty, then its value is split according to
    ``os.pathsep`` and the resulting sequence is passed to
    ``edbob.config.read()``.

    If both ``args`` and ``EDBOB_CONFIG`` are empty, the "standard" locations
    are assumed, and the results of calling both
    :func:`edbob.configuration.default_system_paths()` and
    :func:`edbob.configuration.default_user_paths()` are passed on to
    ``edbob.config.read()``.

    Any other values in ``args`` will be passed directly to
    ``edbob.config.read()`` and so will be interpreted there.  Basically they
    are assumed to be either strings, or sequences of strings, which represent
    paths to various config files, each being read in the order in which it
    appears within ``args``.  (Configuration is thereby cascaded such that the
    file read last will override those before it.)
    """

    config = AppConfigParser(appname)
 
    if args:
        config_paths = list(args)
    elif os.environ.get('EDBOB_CONFIG'):
        config_paths = os.environ['EDBOB_CONFIG'].split(os.pathsep)
    else:
        config_paths = default_system_paths(appname) + default_user_paths(appname)

    service = kwargs.get('service')
    if service:
        config.read_service(service, config_paths)
    else:
        shell = kwargs.get('shell', False)
        for paths in config_paths:
            config.read(paths, recurse=not shell)
    config.configure_logging()

    default_modules = 'edbob.time'
    modules = config.get('edbob', 'init', default=default_modules)
    if modules:
        modules = modules.split(',')
        init_modules(modules, config)

    edbob.graft(edbob, locals(), 'config')
    inited.append('edbob')


def init_modules(names, config=None):
    """
    Initialize the given modules.  ``names`` should be a sequence of strings,
    each of which should be a dotted module name.  If ``config`` is not
    specified, :attr:`edbob.config` is assumed.
    """

    if config is None:
        config = edbob.config

    for name in names:
        name = name.strip()
        if name not in inited:
            module = __import__(name, globals(), locals(), fromlist=['init'])
            if not hasattr(module, 'init'):
                raise InitError(module)
            getattr(module, 'init')(config)
            inited.append(name)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.errors`` -- Error Alert Emails
"""

import sys
import socket
import logging
from traceback import format_exception
from cStringIO import StringIO

import edbob
from edbob.mail import sendmail_with_config


log = logging.getLogger(__name__)


def init(config):
    """
    Creates a system-wide exception hook which logs exceptions and emails them
    to configured recipient(s).
    """

    def excepthook(type, value, traceback):
        email_exception(type, value, traceback)
        sys.__excepthook__(type, value, traceback)

    sys.excepthook = excepthook


def email_exception(type=None, value=None, traceback=None):
    """
    Sends an email containing a traceback to the configured recipient(s).
    """

    if not (type and value and traceback):
        type, value, traceback = sys.exc_info()

    body = StringIO()

    hostname = socket.gethostname()
    body.write("An exception occurred.\n")
    body.write("\n")
    body.write("Machine Name:  %s (%s)\n" % (hostname, socket.gethostbyname(hostname)))
    body.write("Local Time:    %s\n" % (edbob.local_time().strftime('%Y-%m-%d %H:%M:%S %Z%z')))
    body.write("\n")
    body.write("%s\n" % ''.join(format_exception(type, value, traceback)))

    sendmail_with_config('errors', body.getvalue())
    body.close()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.grids.core`` -- Core Grid Classes
"""

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from webhelpers.html import HTML
from webhelpers.html.builder import format_attrs

from pyramid.renderers import render

import edbob


__all__ = ['Grid']


class Grid(edbob.Object):

    full = False
    hoverable = True
    clickable = False
    checkboxes = False
    editable = False
    deletable = False

    partial_only = False

    click_route_name = None
    click_route_kwargs = None

    edit_route_name = None
    edit_route_kwargs = None

    delete_route_name = None
    delete_route_kwargs = None

    def __init__(self, request, **kwargs):
        kwargs.setdefault('fields', OrderedDict())
        kwargs.setdefault('column_titles', {})
        kwargs.setdefault('extra_columns', [])
        super(Grid, self).__init__(**kwargs)
        self.request = request

    def add_column(self, name, label, callback):
        self.extra_columns.append(
            edbob.Object(name=name, label=label, callback=callback))

    def column_header(self, field):
        return HTML.tag('th', field=field.name,
                        title=self.column_titles.get(field.name),
                        c=field.label)

    def div_attrs(self):
        classes = ['grid']
        if self.full:
            classes.append('full')
        if self.clickable:
            classes.append('clickable')
        if self.hoverable:
            classes.append('hoverable')
        return format_attrs(
            class_=' '.join(classes),
            url=self.request.current_route_url())

    def get_delete_url(self, row):
        kwargs = {}
        if self.delete_route_kwargs:
            if callable(self.delete_route_kwargs):
                kwargs = self.delete_route_kwargs(row)
            else:
                kwargs = self.delete_route_kwargs
        return self.request.route_url(self.delete_route_name, **kwargs)

    def get_edit_url(self, row):
        kwargs = {}
        if self.edit_route_kwargs:
            if callable(self.edit_route_kwargs):
                kwargs = self.edit_route_kwargs(row)
            else:
                kwargs = self.edit_route_kwargs
        return self.request.route_url(self.edit_route_name, **kwargs)

    def get_row_attrs(self, row, i):
        attrs = self.row_attrs(row, i)
        if self.clickable:
            kwargs = {}
            if self.click_route_kwargs:
                if callable(self.click_route_kwargs):
                    kwargs = self.click_route_kwargs(row)
                else:
                    kwargs = self.click_route_kwargs
            attrs['url'] = self.request.route_url(self.click_route_name, **kwargs)
        return format_attrs(**attrs)

    def iter_fields(self):
        return self.fields.itervalues()

    def iter_rows(self):
        raise NotImplementedError

    def render(self, template='/grids/grid.mako', **kwargs):
        kwargs.setdefault('grid', self)
        return render(template, kwargs)

    def render_field(self, field):
        raise NotImplementedError

    def row_attrs(self, row, i):
        attrs = {'class_': 'odd' if i % 2 else 'even'}
        return attrs



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.simpleform`` -- pyramid_simpleform Forms
"""

import pyramid_simpleform
from pyramid.renderers import render
from pyramid_simpleform.renderers import FormRenderer

from edbob.pyramid import helpers
from edbob.pyramid.forms import Form


__all__ = ['SimpleForm']


class SimpleForm(Form):

    template = None

    def __init__(self, request, **kwargs):
        super(SimpleForm, self).__init__(request, **kwargs)
        self.form = pyramid_simpleform.Form(request)

    def render(self, **kwargs):
        kw = {
            'form': self,
            }
        kw.update(kwargs)
        return render('/forms/form.mako', kw)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.people`` -- Person Views
"""

from sqlalchemy import and_

# import transaction
# from pyramid.httpexceptions import HTTPFound

# from formalchemy import Field

import edbob
# from edbob.pyramid import filters
# from edbob.pyramid import forms
# from edbob.pyramid import grids
# from edbob.pyramid import Session
from edbob.pyramid.views import SearchableAlchemyGridView, CrudView


class PeopleGrid(SearchableAlchemyGridView):

    mapped_class = edbob.Person
    config_prefix = 'people'
    sort = 'first_name'

    def join_map(self):
        return {
            'email':
                lambda q: q.outerjoin(edbob.PersonEmailAddress, and_(
                    edbob.PersonEmailAddress.parent_uuid == edbob.Person.uuid,
                    edbob.PersonEmailAddress.preference == 1)),
            'phone':
                lambda q: q.outerjoin(edbob.PersonPhoneNumber, and_(
                    edbob.PersonPhoneNumber.parent_uuid == edbob.Person.uuid,
                    edbob.PersonPhoneNumber.preference == 1)),
            }

    def filter_map(self):
        return self.make_filter_map(
            ilike=['first_name', 'last_name'],
            email=self.filter_ilike(edbob.PersonEmailAddress.address),
            phone=self.filter_ilike(edbob.PersonPhoneNumber.number))

    def filter_config(self):
        return self.make_filter_config(
            include_filter_first_name=True,
            filter_type_first_name='lk',
            include_filter_last_name=True,
            filter_type_last_name='lk',
            filter_label_phone="Phone Number",
            filter_label_email="Email Address")

    def sort_map(self):
        return self.make_sort_map(
            'first_name', 'last_name',
            email=self.sorter(edbob.PersonEmailAddress.address),
            phone=self.sorter(edbob.PersonPhoneNumber.number))

    def grid(self):
        g = self.make_grid()
        g.configure(
            include=[
                g.first_name,
                g.last_name,
                g.phone.label("Phone Number"),
                g.email.label("Email Address"),
                ],
            readonly=True)
        g.clickable = True
        g.click_route_name = 'person.read'
        return g


class PersonCrud(CrudView):

    mapped_class = edbob.Person
    home_route = 'people'

    def fieldset(self, model):
        fs = self.make_fieldset(model)
        fs.configure(
            include=[
                fs.first_name,
                fs.last_name,
                fs.phone.label("Phone Number"),
                fs.email.label("Email Address"),
                ])
        return fs


# def filter_map():
#     return filters.get_filter_map(
#         edbob.Person,
#         ilike=['first_name', 'last_name', 'display_name'])

# def search_config(request, fmap):
#     return filters.get_search_config(
#         'people.list', request, fmap,
#         include_filter_display_name=True,
#         filter_type_display_name='lk')

# def search_form(config):
#     return filters.get_search_form(config)

# def grid_config(request, search, fmap):
#     return grids.get_grid_config(
#         'people.list', request, search,
#         filter_map=fmap, sort='display_name')

# def sort_map():
#     return grids.get_sort_map(
#         edbob.Person,
#         ['first_name', 'last_name', 'display_name'])

# def query(config):
#     smap = sort_map()
#     q = Session.query(edbob.Person)
#     q = filters.filter_query(q, config)
#     q = grids.sort_query(q, config, smap)
#     return q


# def people(context, request):

#     fmap = filter_map()
#     config = search_config(request, fmap)
#     search = search_form(config)
#     config = grid_config(request, search, fmap)
#     people = grids.get_pager(query, config)

#     g = forms.AlchemyGrid(
#         edbob.Person, people, config,
#         gridurl=request.route_url('people.list'),
#         objurl='person.edit')

#     g.configure(
#         include=[
#             g.first_name,
#             g.last_name,
#             g.display_name,
#             ],
#         readonly=True)

#     grid = g.render(class_='clickable people')
#     return grids.render_grid(request, grid, search)


# def person_fieldset(person, request):
#     fs = forms.make_fieldset(person, url=request.route_url,
#                              url_action=request.current_route_url(),
#                              route_name='people.list')
#     fs.configure(
#         include=[
#             fs.first_name,
#             fs.last_name,
#             fs.display_name,
#             ])
#     return fs


# def new_person(context, request):

#     fs = person_fieldset(edbob.Person, request)
#     if not fs.readonly and request.POST:
#         fs.rebind(data=request.params)
#         if fs.validate():

#             with transaction.manager:
#                 fs.sync()
#                 Session.add(fs.model)
#                 Session.flush()
#                 request.session.flash("%s \"%s\" has been %s." % (
#                         fs.crud_title, fs.get_display_text(),
#                         'updated' if fs.edit else 'created'))

#             return HTTPFound(location=request.route_url('people.list'))

#     return {'fieldset': fs, 'crud': True}


# def edit_person(request):
#     """
#     View for editing a :class:`edbob.Person` instance.
#     """

#     from edbob.pyramid.views.users import user_fieldset

#     uuid = request.matchdict['uuid']
#     person = Session.query(edbob.Person).get(uuid) if uuid else None
#     assert person

#     fs = person_fieldset(person, request)
#     if request.POST:
#         fs.rebind(data=request.params)
#         if fs.validate():

#             with transaction.manager:
#                 fs.sync()
#                 fs.model = Session.merge(fs.model)
#                 request.session.flash("%s \"%s\" has been %s." % (
#                         fs.crud_title, fs.get_display_text(),
#                         'updated' if fs.edit else 'created'))
#                 home = request.route_url('people.list')

#             return HTTPFound(location=home)

#     user = fs.model.user
#     if user:
#         user = user_fieldset(user, request)
#         user.readonly = True
#         del user.person
#         del user.password
#         del user.confirm_password

#     return {'fieldset': fs, 'crud': True, 'user': user}


def includeme(config):

    config.add_route('people', '/people')
    config.add_view(PeopleGrid, route_name='people',
                    renderer='/people/index.mako',
                    permission='people.list')

#     config.add_route('people.list', '/people')
#     config.add_view(people, route_name='people.list', renderer='/people/index.mako',
#                     permission='people.list', http_cache=0)

    config.add_route('person.read', '/people/{uuid}')
    config.add_view(PersonCrud, attr='read', route_name='person.read',
                    renderer='/people/crud.mako',
                    permission='people.read')

#     config.add_route('person.new', '/people/new')
#     config.add_view(new_person, route_name='person.new', renderer='/people/person.mako',
#                     permission='people.create', http_cache=0)

#     config.add_route('person.edit', '/people/{uuid}/edit')
#     config.add_view(edit_person, route_name='person.edit', renderer='/people/person.mako',
#                     permission='people.edit', http_cache=0)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.roles`` -- Role Views
"""

import transaction
from pyramid.httpexceptions import HTTPFound

from formalchemy import Field, FieldRenderer
from webhelpers.html import literal
from webhelpers.html.tags import checkbox, hidden

import edbob
from edbob.db.auth import administrator_role, has_permission
from edbob.pyramid import filters
from edbob.pyramid import forms
from edbob.pyramid import grids
from edbob.pyramid import Session


def filter_map():
    return filters.get_filter_map(
        edbob.Role,
        ilike=['name'])

def search_config(request, fmap):
    return filters.get_search_config(
        'roles.list', request, fmap,
        include_filter_name=True,
        filter_type_name='lk')

def search_form(config):
    return filters.get_search_form(config)

def grid_config(request, search, fmap):
    return grids.get_grid_config(
        'roles.list', request, search,
        filter_map=fmap, sort='name')

def sort_map():
    return grids.get_sort_map(edbob.Role, ['name'])

def query(config):
    smap = sort_map()
    q = Session.query(edbob.Role)
    q = filters.filter_query(q, config)
    q = grids.sort_query(q, config, smap)
    return q


def roles(request):

    fmap = filter_map()
    config = search_config(request, fmap)
    search = search_form(config)
    config = grid_config(request, search, fmap)
    roles = grids.get_pager(query, config)

    g = forms.AlchemyGrid(
        edbob.Role, roles, config,
        gridurl=request.route_url('roles.list'),
        objurl='role.edit')

    g.configure(
        include=[
            g.name,
            ],
        readonly=True)

    grid = g.render(class_='clickable roles')
    return grids.render_grid(request, grid, search)


class PermissionsField(Field):

    def sync(self):
        if not self.is_readonly():
            role = self.model
            role.permissions = self.renderer.deserialize()


class PermissionsFieldRenderer(FieldRenderer):

    available_permissions = [

        ("Batches", [
                ('batches.list',        "List Batches"),
                ('batches.edit',        "Edit Batch"),
                ('batches.create',      "Create Batch"),
                ]),

        ("Roles", [
                ('roles.list',          "List Roles"),
                ('roles.edit',          "Edit Role"),
                ('roles.create',        "Create Role"),
                ]),
        ]

    def deserialize(self):
        perms = []
        i = len(self.name) + 1
        for key in self.params:
            if key.startswith(self.name):
                perms.append(key[i:])
        return perms

    def _render(self, readonly=False, **kwargs):
        # result = literal('')
        # for group_name, group_label, perm_list in self.field.model_value:
        #     rendered_group_name = literal('<p class="permission-group">' + group_label + '</p>\n')
        #     if readonly:
        #         result += literal('<tr><td colspan="2">') + rendered_group_name + literal('</td></tr>')
        #     else:
        #         result += rendered_group_name
        #         result += literal('<div>')
        #     for perm_name, perm_label, checked in perm_list:
        #         if readonly:
        #             result += literal('<tr>'
        #                               + '<td class="permission">' + ('[X]' if checked else '[&nbsp; ]') + '</td>'
        #                               + '<td class="permission-label">' + perm_label + '</td>'
        #                               + '</tr>\n')
        #         else:
        #             name = '.'.join((self.name, group_name, perm_name))
        #             result += check_box(name, label=perm_label, checked=checked)
        #     if not readonly:
        #         result += literal('</div>')
        # if readonly:
        #     return literal('<table class="permissions">') + result + literal('</table>')
        # return literal('<div class="permissions">') + result + literal('</div>')

        role = self.field.model
        if role is administrator_role(Session()):
            res = literal('<p>This is the administrative role; '
                          'it has full access to the entire system.</p>')
            if not readonly:
                res += hidden(self.name, value='') # ugly hack..or good idea?
        else:
            res = ''
            for group, perms in self.available_permissions:
                res += literal('<p class="group">%s</p>' % group)
                for perm, title in perms:
                    if readonly:
                        res += literal('<p>%s</p>' % title)
                    else:
                        checked = has_permission(role, perm)
                        res += checkbox(self.name + '-' + perm,
                                        checked=checked, label=title)
        return res

    def render(self, **kwargs):
        return self._render(**kwargs)

    def render_readonly(self, **kwargs):
        return self._render(readonly=True, **kwargs)


def role_fieldset(role, request):
    fs = forms.make_fieldset(role, url=request.route_url,
                             url_action=request.current_route_url(),
                             route_name='roles.list')
    
    fs.append(PermissionsField('permissions',
                               renderer=PermissionsFieldRenderer))

    fs.configure(
        include=[
            fs.name,
            fs.permissions,
            ])

    if not fs.edit:
        del fs.permissions

    return fs


def new_role(request):

    fs = role_fieldset(edbob.Role, request)
    if request.POST:
        fs.rebind(data=request.params)
        if fs.validate():

            with transaction.manager:
                fs.sync()
                fs.model = Session.merge(fs.model)
                request.session.flash("%s \"%s\" has been %s." % (
                        fs.crud_title, fs.get_display_text(),
                        'updated' if fs.edit else 'created'))
                home = request.route_url('roles.list')

            return HTTPFound(location=home)

    return {'fieldset': fs, 'crud': True}


def edit_role(request):
    uuid = request.matchdict['uuid']
    role = Session.query(edbob.Role).get(uuid) if uuid else None
    assert role

    fs = role_fieldset(role, request)
    if request.POST:
        fs.rebind(data=request.params)
        if fs.validate():

            with transaction.manager:
                Session.add(fs.model)
                fs.sync()
                request.session.flash("%s \"%s\" has been %s." % (
                        fs.crud_title, fs.get_display_text(),
                        'updated' if fs.edit else 'created'))
                home = request.route_url('roles.list')

            return HTTPFound(location=home)

    return {'fieldset': fs, 'crud': True}


def includeme(config):

    config.add_route('roles.list', '/roles')
    config.add_view(roles, route_name='roles.list', renderer='/roles/index.mako',
                    permission='roles.list', http_cache=0)

    config.add_route('role.new', '/roles/new')
    config.add_view(new_role, route_name='role.new', renderer='/roles/role.mako',
                    permission='roles.create', http_cache=0)

    config.add_route('role.edit', '/roles/{uuid}/edit')
    config.add_view(edit_role, route_name='role.edit', renderer='/roles/role.mako',
                    permission='roles.edit', http_cache=0)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.filemon`` -- File Monitoring Service
"""

import os.path
import logging

import edbob


log = logging.getLogger(__name__)


class MonitorProfile(object):
    """
    This is a simple profile class, used to represent configuration of the file
    monitor service.
    """

    def __init__(self, appname, key):
        self.appname = appname
        self.key = key

        self.dirs = edbob.config.require('%s.filemon' % appname, '%s.dirs' % key)
        self.dirs = eval(self.dirs)

        actions = edbob.config.require('%s.filemon' % appname, '%s.actions' % key)
        actions = eval(actions)

        self.actions = []
        for action in actions:
            if isinstance(action, tuple):
                spec = action[0]
                args = list(action[1:])
            else:
                spec = action
                args = []
            func = edbob.load_spec(spec)
            self.actions.append((spec, func, args))

        self.locks = edbob.config.getboolean(
            '%s.filemon' % appname, '%s.locks' % key, default=False)


def get_monitor_profiles(appname):
    """
    Convenience function to load monitor profiles from config.
    """

    monitored = {}

    # Read monitor profile(s) from config.
    keys = edbob.config.require('%s.filemon' % appname, 'monitored')
    keys = keys.split(',')
    for key in keys:
        key = key.strip()
        profile = MonitorProfile(appname, key)
        monitored[key] = profile
        for path in profile.dirs[:]:

            # Ensure the monitored path exists.
            if not os.path.exists(path):
                log.warning("get_monitor_profiles: Profile '%s' has nonexistent "
                            "path, which will be pruned: %s" % (key, path))
                profile.dirs.remove(path)

            # Ensure the monitored path is a folder.
            elif not os.path.isdir(path):
                log.warning("get_monitor_profiles: Profile '%s' has non-folder "
                            "path, which will be pruned: %s" % (key, path))
                profile.dirs.remove(path)

    for key in monitored.keys():
        profile = monitored[key]

        # Prune any profiles with no valid folders to monitor.
        if not profile.dirs:
            log.warning("get_monitor_profiles: Profile '%s' has no folders to "
                        "monitor, and will be pruned." % key)
            del monitored[key]

        # Prune any profiles with no valid actions to perform.
        elif not profile.actions:
            log.warning("get_monitor_profiles: Profile '%s' has no actions to "
                        "perform, and will be pruned." % key)
            del monitored[key]

    return monitored



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.wx`` -- wxPython Framework
"""

from __future__ import absolute_import

import wx


class ProgressDialog(wx.ProgressDialog):
    
    def __init__(self, parent, message, maximum, title="Processing...", can_abort=True, *args, **kwargs):
        style = wx.PD_SMOOTH|wx.PD_AUTO_HIDE|wx.PD_APP_MODAL|wx.PD_ELAPSED_TIME|wx.PD_REMAINING_TIME
        if can_abort:
            style |= wx.PD_CAN_ABORT
        if 'style' in kwargs:
            style &= kwargs['style']
        kwargs['style'] = style
        wx.ProgressDialog.__init__(self, title, message, maximum=maximum, parent=parent, *args, **kwargs)
        
    def update(self, value, *args, **kwargs):
        if not wx.ProgressDialog.Update(self, value, *args, **kwargs)[0]:
            if self.ConfirmAbort():
                return False
            self.Resume()
        return True
        
    def destroy(self):
        self.Destroy()
        
    def ConfirmAbort(self):
        dlg = wx.MessageDialog(self, "Do you really wish to cancel this process?",
                               "Really Cancel?", wx.ICON_QUESTION|wx.YES_NO|wx.NO_DEFAULT)
        res = dlg.ShowModal()
        dlg.Destroy()
        return res == wx.ID_YES

        
class ProgressFactory(object):

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        self.args = args
        self.kwargs = kwargs

    def __call__(self, message, maximum, *args, **kwargs):
        message = '%s ...' % message
        args = self.args + args
        _kwargs = self.kwargs.copy()
        _kwargs.update(kwargs)
        return ProgressDialog(self.parent, message, maximum, *args, **_kwargs)


def LaunchDialog(dialog_class):
    """
    Creates a ``wx.PySimpleApp``, then instantiates ``dialog_class`` and shows
    it modally.
    """
    app = wx.PySimpleApp()
    dlg = dialog_class(None)
    dlg.ShowModal()
    dlg.Destroy()
    app.MainLoop()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.exceptions`` -- Database Exceptions
"""


class CoreSchemaAlreadyInstalled(Exception):
    """
    Raised when a request is made to install the core schema to a database, but
    it is already installed there.
    """

    def __init__(self, installed_version):
        self.installed_version = installed_version

    def __str__(self):
        return "Core schema already installed (version %s)" % self.installed_version


class CoreSchemaNotInstalled(Exception):
    """
    Raised when a request is made which requires the core schema to be present
    in a database, yet such is not the case.
    """

    def __init__(self, engine):
        self.engine = engine

    def __str__(self):
        return "Core schema not installed: %s" % str(self.engine)


class ExtensionNotFound(Exception):
    """
    Raised when an extension is requested which cannot be located.
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "Extension not found: %s" % self.name



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.classes`` -- Data Class Definitions
"""

from sqlalchemy.ext.associationproxy import association_proxy

import edbob
from edbob.sqlalchemy import getset_factory


__all__ = ['Permission', 'Person', 'Role', 'Setting', 'User', 'UserRole']


class PersonDerivative(edbob.Object):
    """
    Base class for classes which must derive certain functionality from the
    :class:`Person` class, e.g. :class:`User`.
    """

    display_name = association_proxy('person', 'display_name',
                                     creator=lambda x: Person(display_name=x),
                                     getset_factory=getset_factory)

    first_name = association_proxy('person', 'first_name',
                                   creator=lambda x: Person(first_name=x),
                                   getset_factory=getset_factory)

    last_name = association_proxy('person', 'last_name',
                                  creator=lambda x: Person(last_name=x),
                                  getset_factory=getset_factory)


class ActiveExtension(edbob.Object):
    """
    Represents an extension which has been activated within a database.
    """

    def __repr__(self):
        return "<ActiveExtension: %s>" % self.name


class Permission(edbob.Object):
    """
    Represents the fact that a particular :class:`Role` is allowed to do a
    certain thing.
    """

    def __repr__(self):
        return "<Permission: %s: %s>" % (self.role, self.permission)


class Person(edbob.Object):
    """
    Represents a real, living and breathing person.  (Or, at least was
    previously living and breathing, in the case of the deceased.)
    """

    def __repr__(self):
        return "<Person: %s>" % self.display_name

    def __str__(self):
        return str(self.display_name or '')

    @property
    def customer(self):
        """
        Returns the first :class:`Customer` instance in
        :attr:`Person.customers`, or ``None`` if that list is empty.

        .. note::
           As of this writing, :attr:`Person.customers` is an
           arbitrarily-ordered list, so the only real certainty you have when
           using :attr:`Person.customer` is when the :class:`Person` instance
           is associated with exactly one (or zero) :class:`Customer`
           instances.
        """

        if self.customers:
            return self.customers[0]
        return None


class Role(edbob.Object):
    """
    Represents a role within the organization; used to manage permissions.
    """

    permissions = association_proxy('_permissions', 'permission',
                                    creator=lambda x: Permission(permission=x),
                                    getset_factory=getset_factory)

    users = association_proxy('_users', 'user',
                              creator=lambda x: UserRole(user=x),
                              getset_factory=getset_factory)

    def __repr__(self):
        return "<Role: %s>" % self.name

    def __str__(self):
        return str(self.name or '')


class Setting(edbob.Object):
    """
    Represents a setting stored within the database.
    """

    def __repr__(self):
        return "<Setting: %s>" % self.name


class User(PersonDerivative):
    """
    Represents a user of the system.  This may or may not correspond to a real
    person, e.g. for data import jobs and the like.
    """

    employee = association_proxy('person', 'employee',
                                 creator=lambda x: Person(employee=x),
                                 getset_factory=getset_factory)

    roles = association_proxy('_roles', 'role',
                              creator=lambda x: UserRole(role=x),
                              getset_factory=getset_factory)

    def __repr__(self):
        return "<User: %s>" % self.username

    def __str__(self):
        return str(self.username or '')


class UserRole(edbob.Object):
    """
    Represents the association between a :class:`User` and a :class:`Role`.
    """

    def __repr__(self):
        return "<UserRole: %s : %s>" % (self.user, self.role)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################


try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()


import sys
import os.path
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
execfile(os.path.join(here, 'edbob', '_version.py'))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()


requires = [
    #
    # Version numbers within comments below have specific meanings.
    # Basically the 'low' value is a "soft low," and 'high' a "soft high."
    # In other words:
    #
    # If either a 'low' or 'high' value exists, the primary point to be
    # made about the value is that it represents the most current (stable)
    # version available for the package (assuming typical public access
    # methods) whenever this project was started and/or documented.
    # Therefore:
    #
    # If a 'low' version is present, you should know that attempts to use
    # versions of the package significantly older than the 'low' version
    # may not yield happy results.  (A "hard" high limit may or may not be
    # indicated by a true version requirement.)
    #
    # Similarly, if a 'high' version is present, and especially if this
    # project has laid dormant for a while, you may need to refactor a bit
    # when attempting to support a more recent version of the package.  (A
    # "hard" low limit should be indicated by a true version requirement
    # when a 'high' version is present.)
    #
    # In any case, developers and other users are encouraged to play
    # outside the lines with regard to these soft limits.  If bugs are
    # encountered then they should be filed as such.
    #
    # package                           # low                   high

    'decorator',                        # 3.3.2
    'lockfile',                         # 0.9.1
    'progressbar',                      # 2.3
    'pytz',                             # 2012b
    ]

if sys.version_info < (2, 7):
    # Python < 2.7 has a standard library in need of supplementation.

    requires += [
        #
        # package                       # low                   high

        'argparse',                     # 1.2.1
        'ordereddict',                  # 1.1
        ]

    
extras = {

    'db': [
        #
        # package                       # low                   high

        'alembic',                      # 0.3.1
        'SQLAlchemy',                   # 0.7.6
        # 'Tempita',                      # 0.5.1
        ],

    'docs': [
        #
        # package                       # low                   high

        'Sphinx',                       # 1.1.3
        ],

    'filemon': [
        #
        # package                       # low                   high

        # This is just a placeholder on Windows; Linux requires this extra.
        ],

    'pyramid': [
        #
        # package                       # low                   high

        # Beaker dependency included here because 'pyramid_beaker' uses incorrect
        # case in its requirement declaration.
        'Beaker',                       # 1.6.3

        # Pyramid 1.3 introduced 'pcreate' command (and friends) to replace
        # deprecated 'paster create' (and friends).
        'pyramid>=1.3a1',               #                       1.3b2

        'FormAlchemy',                  # 1.4.2
        'FormEncode',                   # 1.2.4
        'Mako',                         # 0.6.2
        'pyramid_beaker>=0.6',          #                       0.6.1
        'pyramid_debugtoolbar',         # 1.0
        'pyramid_exclog',               # 0.6
        'pyramid_simpleform',           # 0.6.1
        'pyramid_tm',                   # 0.3
        # 'Tempita',                      # 0.5.1
        'transaction',                  # 1.2.0
        'waitress',                     # 0.8.1
        'WebHelpers',                   # 1.3
        'zope.sqlalchemy',              # 0.7
        ],
    }

if sys.platform == 'win32':

    extras['db'] += [
        #
        # package                       # low                   high

        'py-bcrypt-w32',                # 0.2.2
        ]

    extras['pyramid'] += [
        #
        # package                       # low                   high

        'py-bcrypt-w32',                # 0.2.2
        ]

elif sys.platform == 'linux2':

    extras['db'] += [
        #
        # package                       # low                   high

        'py-bcrypt',                    # 0.2
        ]

    extras['filemon'] += [
        #
        # package                       # low                   high

        'pyinotify',                    # 0.9.3
        ]

    extras['pyramid'] += [
        #
        # package                       # low                   high

        'py-bcrypt',                    # 0.2
        ]


setup(
    name = "edbob",
    version = __version__,
    author = "Lance Edgar",
    author_email = "lance@edbob.org",
    url = "http://edbob.org/",
    license = "GNU Affero GPL v3",
    description = "Pythonic Software Framework",
    long_description = README + '\n\n' +  CHANGES,

    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],

    install_requires = requires,
    extras_require = extras,

    packages = find_packages(),
    include_package_data = True,
    zip_safe = False,

    entry_points = """

[console_scripts]
edbob = edbob.commands:main

[gui_scripts]
edbobw = edbob.commands:main

[pyramid.scaffold]
edbob = edbob.scaffolds:Template

[edbob.commands]
db = edbob.commands:DatabaseCommand
filemon = edbob.commands:FileMonitorCommand
shell = edbob.commands:ShellCommand
uuid = edbob.commands:UuidCommand

[edbob.db.extensions]
auth = edbob.db.extensions.auth:AuthExtension
contact = edbob.db.extensions.contact:ContactExtension

""",
    )



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.auth`` -- Auth Views
"""

import formencode
from pyramid.httpexceptions import HTTPFound
from pyramid.security import remember, forget
from pyramid_simpleform import Form
from pyramid_simpleform.renderers import FormRenderer

import edbob
from edbob.db.auth import authenticate_user
from edbob.pyramid import Session


class UserLogin(formencode.Schema):
    allow_extra_fields = True
    filter_extra_fields = True
    username = formencode.validators.NotEmpty()
    password = formencode.validators.NotEmpty()


def login(request):
    """
    The login view, responsible for displaying and handling the login form.
    """

    referrer = request.get_referrer()

    # Redirect if already logged in.
    if request.user:
        return HTTPFound(location=referrer)

    form = Form(request, schema=UserLogin)
    if form.validate():
        user = authenticate_user(form.data['username'],
                                 form.data['password'],
                                 session=Session())
        if user:
            request.session.flash("%s logged in at %s" % (
                    user.display_name,
                    edbob.local_time().strftime('%I:%M %p')))
            headers = remember(request, user.uuid)
            return HTTPFound(location=referrer, headers=headers)
        request.session.flash("Invalid username or password")

    url = edbob.config.get('edbob.pyramid', 'login.logo_url',
                           default=request.static_url('edbob.pyramid:static/img/logo.jpg'))
    kwargs = eval(edbob.config.get('edbob.pyramid', 'login.logo_kwargs',
                                   default="dict(width=500)"))

    return {'form': FormRenderer(form), 'referrer': referrer,
            'logo_url': url, 'logo_kwargs': kwargs}


def logout(request):
    """
    View responsible for logging out the current user.

    This deletes/invalidates the current session and then redirects to the
    login page.
    """

    request.session.delete()
    request.session.invalidate()
    headers = forget(request)
    referrer = request.get_referrer()
    return HTTPFound(location=referrer, headers=headers)


def includeme(config):

    config.add_route('login', '/login')
    config.add_view(login, route_name='login', renderer='/login.mako')

    config.add_route('logout', '/logout')
    config.add_view(logout, route_name='logout')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.files`` -- Files & Folders
"""


import os
import os.path
import shutil
import tempfile
import lockfile


__all__ = ['temp_path']


class DosFile(file):
    """
    Subclass of ``file`` which uses DOS line endings when writing the file.
    """

    def write(self, string):
        super(DosFile, self).write(string.replace(os.linesep, '\r\n'))


def locking_copy(src, dst):
    """
    Implements a "locking" version of ``shutil.copy()``.

    This function exists to provide a more atomic method for copying a file
    into a Linux folder which is being watched by a file monitor running on the
    Linux machine.  This is necessary because it is not practical to watch for
    any particular ``pyinotify`` event in order to know when the file is "free"
    - at least in the case of a simple copy.  The reason for this is that
    ``shutil.copy()`` first copies the file, but then will attempt to change
    its attributes.  Under normal circumstances it would seem best to respond
    to the "create" (or "write close") event on the file, but in this case the
    attribute update really must occur before the watched file is processed.
    """

    fn = os.path.basename(src)
    dst = os.path.join(dst, fn)
    with lockfile.FileLock(dst):
        shutil.copy(src, dst)


def change_newlines(path, newline):
    """
    Rewrites the file at ``path``, changing its newline character(s) to that of
    ``newline``.
    """

    root, ext = os.path.splitext(path)
    temp_path = temp_path(suffix='.' + ext)
    infile = open(path, 'rUb')
    outfile = open(temp_path, 'wb')
    for line in infile:
        line = line.rstrip('\r\n')
        outfile.write(line + newline)
    infile.close()
    outfile.close()
    os.remove(path)
    shutil.move(temp_path, path)


def count_lines(path):
    """
    Convenience function to count the number of lines in a text file.  Some
    attempt is made to ensure cross-platform compatibility.
    """

    f = open(path, 'rb')
    lines = f.read().count('\n') + 1
    f.close()
    return lines


def temp_path(suffix='.tmp', prefix='edbob.'):
    """
    Convenience function to return a temporary file path.  The arguments'
    meanings are the same as for ``tempfile.mkstemp()``.
    """

    fd, path = tempfile.mkstemp(suffix=suffix, prefix=prefix)
    os.close(fd)
    os.remove(path)
    return path



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.core`` -- Core Forms
"""

from sqlalchemy.util import OrderedDict

from webhelpers.html import literal, tags

import edbob
from edbob.util import requires_impl


__all__ = ['Form']


class Form(edbob.Object):
    """
    Generic form class.

    This class exists primarily so that rendering calls may mimic those used by
    FormAlchemy.
    """

    readonly = False
    successive = False

    action_url = None
    home_route = None
    home_url = None
    # template = None

    render_fields = OrderedDict()
    errors = {}

    # def __init__(self, request=None, action_url=None, home_url=None, template=None, **kwargs):
    def __init__(self, request=None, action_url=None, home_url=None, **kwargs):
        super(Form, self).__init__(**kwargs)
        self.request = request
        if action_url:
            self.action_url = action_url
        if request and not self.action_url:
            self.action_url = request.current_route_url()
        if home_url:
            self.home_url = home_url
        if request and not self.home_url:
            home = self.home_route if self.home_route else 'home'
            self.home_url = request.route_url(home)
        # if template:
        #     self.template = template
        # if not self.template:
        #     self.template = '%s.mako' % self.action_url

    @property
    def action_url(self):
        return self.request.current_route_url()

    def standard_buttons(self, submit="Save"):
        return literal(tags.submit('submit', submit) + ' ' + self.cancel_button())

    def cancel_button(self):
        return literal('<button type="button" class="cancel">Cancel</button>')

    def render(self, **kwargs):
        """
        Renders the form as HTML.  All keyword arguments are passed on to the
        template context.
        """

        return ''



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.progress`` -- Progress Indicator
"""

from beaker.session import Session


def get_progress_session(session, key):
    request = session.request
    id = '%s.progress.%s' % (session.id, key)
    session = Session(request, id)
    return session


class SessionProgress(object):
    """
    Provides a session-based progress bar mechanism.

    This class is only responsible for keeping the progress *data* current.  It
    is the responsibility of some client-side AJAX (etc.) to consume the data
    for display to the user.
    """

    def __init__(self, session, key):
        self.session = get_progress_session(session, key)
        self.cancelled = False

    def __call__(self, message, maximum):
        self.session['complete'] = False
        self.session['message'] = message
        self.session['maximum'] = maximum
        self.session['cancelled'] = False
        self.session['value'] = 0
        self.session.save()
        return self

    def update(self, value):
        self.session.load()
        if self.session.get('cancelled'):
            self.cancelled = True
        else:
            self.session['value'] = value
            self.session.save()
        return not self.cancelled

    def destroy(self):
        if not self.cancelled:
            self.session['complete'] = True
            self.session.save()

    def secondary_progress(self):
        return SecondarySessionProgress(self)
    

class SecondarySessionProgress(object):

    def __init__(self, parent):
        self.parent = parent
        self.session = parent.session

    def __call__(self, message, maximum):
        self.session['message'] = message
        self.session['value'] = 0
        self.session['maximum'] = maximum
        self.session.save()
        return self

    def update(self, value):
        return self.parent.update(value)

    def destroy(self):
        pass
